@@ -5,6 +5,7 @@ from mixsol.helpers import name_to_components, components_to_name, calculate_mol
 from mixsol.digraph import DirectedGraph
 import itertools as itt
 import matplotlib.pyplot as plt
+from random import sample
 
 
 class Mixer:
@@ -58,8 +59,12 @@ class Mixer:
         """
         if solution_indices is None:
             solution_indices = list(range(len(self.solutions)))
-        A = self.solution_matrix[solution_indices]
-        b = self._solution_to_vector(target)
+        A0 = self.solution_matrix[solution_indices]
+        scale = A0[A0 > 0].min()  # smallest nonzero component
+        A = (
+            A0 / scale
+        )  # change min value to 1, avoids rounding errors with trace components in Solution
+        b = self._solution_to_vector(target) / scale
         if not self._is_plausible(A, b):
             return np.nan
         x, err = nnls(A.T, b, maxiter=1e3)
@@ -101,7 +106,7 @@ class Mixer:
                 target_matrix[m] = target.solvent[c] * volume
         return target_matrix.T
 
-    def mix(
+    def _mix_to_vector(
         self,
         target: Solution,
         volume: float,
@@ -129,9 +134,9 @@ class Mixer:
         Returns:
             np.ndarray: vector of solution volumes corresponding to rows in the solution matrix (self.solutions). If verbose=True, this will be a list of such vectors that all reach the target solution
         """
-        if strategy not in ["least_inputs", "prefer_stock"]:
+        if strategy not in ["least_inputs", "prefer_stock", "fastest"]:
             raise ValueError(
-                "Mixing strategy must be either 'least_inputs' or 'prefer_stock'"
+                "Mixing strategy must be 'least_inputs', 'prefer_stock', or 'fastest'"
             )
         min_fraction = min_volume / volume
         if solution_indices is None or solution_indices == "stock":
@@ -141,9 +146,15 @@ class Mixer:
 
         possible_mixtures = []
         if max_inputs is None:
-            max_inputs = len(self.solutions)
+            max_inputs = len(solution_indices)
+        max_inputs = min(
+            volume // min_volume, max_inputs
+        )  # min pipette volume may limit us to fewer inputs
+
         for i in range(1, max_inputs):
-            for idx in itt.combinations(solution_indices, i):
+            for idx in itt.combinations(
+                sample(solution_indices, len(solution_indices)), i
+            ):
                 x = self._calculate_mix(
                     target,
                     solution_indices=list(idx),
@@ -154,6 +165,10 @@ class Mixer:
                     possible_mixtures.append(
                         x.round(12) * volume
                     )  # remove floating pt precision errors
+                    if strategy == "fastest":
+                        break
+            if len(possible_mixtures) > 0:
+                break
         if len(possible_mixtures) == 0:
             return np.nan
         possible_mixtures = [pm for pm in np.unique(possible_mixtures, axis=0)]
@@ -217,7 +232,7 @@ class Mixer:
             for i in range(len(self.solutions)):
                 if i in self.availableidx:
                     continue
-                x = self.mix(
+                x = self._mix_to_vector(
                     target=self.solutions[i],
                     volume=self.target_volumes[self.solutions[i]],
                     solution_indices=self.availableidx,
@@ -235,6 +250,64 @@ class Mixer:
         self.__solved = True
         return graph
 
+    def mix(
+        self,
+        target: Solution,
+        volume: float,
+        solutions_to_use: list = "stock",
+        tolerance: float = 1e-5,
+        min_volume: float = 0,
+        verbose: bool = False,
+        max_inputs: int = None,
+        strategy: str = "least_inputs",
+    ) -> dict:
+        """Calculate mixture of stock solutions to achieve target solution
+
+        Args:
+            target (Solution): target solution
+            volume (float): volume of target solution desired
+            solutions_to_use (str, optional): "stock" or "all" to use stock solutions or all (stocks + targets) solutions as potential inputs for the mixture. Defaults to "stock".
+            tolerance (float, optional): error threhold for target solution. Defaults to 1e-2.
+            min_volume (float, optional): minimum volume that can be mixed from any single solution (useful for pipettes with a minimum aspiration volume). Defaults to 0.
+            verbose (bool, optional): if True, returns all plausible mixture vectors. If False (default), only returns the best (largest minimum single volume transfer) vector.
+            max_inputs (int, optional): maximum number of solutions to mix into the given target. Defaults to None (no limit).
+            strategy (str, optional): strategy to select mixing inputs from the set of valid inputs.
+                "least_inputs": select mixing inputs such that the smallest input volume is maximized. this should mix with the least number of input solutions
+                "prefer_stock": select mixing inputs such that the number of non-stock inputs is minimized.
+
+        Returns:
+            np.ndarray: vector of solution volumes corresponding to rows in the solution matrix (self.solutions). If verbose=True, this will be a list of such vectors that all reach the target solution
+        """
+        if strategy not in ["least_inputs", "prefer_stock", "fastest"]:
+            raise ValueError(
+                "Mixing strategy must be 'least_inputs', 'prefer_stock', or 'fastest'"
+            )
+        if solutions_to_use == "stock":
+            solution_indices = self.stock_idx
+        elif solutions_to_use == "all":
+            solution_indices = list(range(len(self.solutions)))
+        else:
+            raise ValueError("solutions_to_use must be 'stock' or 'all'")
+
+        volume_vector = self._mix_to_vector(
+            target=target,
+            volume=volume,
+            tolerance=tolerance,
+            min_volume=min_volume,
+            solution_indices=solution_indices,
+            verbose=verbose,
+            strategy=strategy,
+            max_inputs=max_inputs,
+        )
+        if volume_vector is np.nan:
+            raise Exception(f"Could not find a valid set of volume transfers to reach the target solution {target}!")
+        volumes_dict = {
+            self.solutions[i]: volume
+            for i, volume in enumerate(volume_vector)
+            if volume > 0
+        }
+        return volumes_dict
+
     def solve(
         self,
         min_volume: float,
@@ -305,6 +378,65 @@ class Mixer:
             if len(this_gen) > 0:
                 self.transfers_per_generation.append(this_gen)
 
+    def all_possible_solutions(
+        self,
+        min_volume: float,
+        target_volume: float,
+        precision: float,
+        max_inputs: int = 4,
+    ):
+        """Constructs a list of all Solution's that can be accessed from the set of stock solutions. Note that all generated solutions will have a molarity of one (doesn't change the actual contents, just the labeling, molarity is reflected in the solutes coefficients).
+
+        Args:
+            min_volume (float): minimum transfer volume
+            target_volume (float): final volume of accessible Solution
+            precision (float): increment of volume at which to mix stocks
+            max_inputs (int, optional): Max number of stock solutinos that can be mixed. This may be overridden if min_volume/total_volume limits the number of inputs. Defaults to 4.
+
+        Returns:
+            list: list of accessible Solution's
+        """
+        max_inputs = min(
+            max_inputs, int(np.floor(target_volume / min_volume))
+        )  # possible that we are limited by volumes vs user guidance
+
+        # ratios = self.__all_possible_ratios(
+        #     num_inputs=max_inputs,
+        #     min_fraction=min_volume / target_volume,
+        #     steps=int(target_volume / precision),
+        # )
+        num_steps = int(target_volume / precision)
+        ratios = []
+        for c in itt.combinations_with_replacement(list(range(max_inputs)), num_steps):
+            this_ratio = [c.count(i) / num_steps for i in range(max_inputs)]
+            if min([r for r in this_ratio if r > 0]) < (min_volume / target_volume):
+                continue
+            ratios.append(this_ratio)
+
+        solution_vectors = []
+        for s in itt.combinations(self.solution_matrix, max_inputs):
+            for r in ratios:
+                row = np.array(s).T @ r
+                solution_vectors.append(row)
+        solution_vectors = np.array(solution_vectors).round(12)
+        solution_vectors = np.unique(solution_vectors, axis=0)
+
+        solutions = []
+        for r in solution_vectors:
+            solutes = {
+                c: amt
+                for idx, (c, amt) in enumerate(zip(self.components, r))
+                if idx not in self.solvent_idx
+            }
+            solvents = {
+                c: amt
+                for idx, (c, amt) in enumerate(zip(self.components, r))
+                if idx in self.solvent_idx
+            }
+            solutions.append(Solution(solutes=solutes, solvent=solvents, molarity=1))
+
+        return solutions
+
     ### publishing methods
     def _check_if_solved(self):
         if not self.__solved:
@@ -462,11 +594,14 @@ class Weigher:
         Returns:
             list: row indices of powders that are valid options for the target solution
         """
-        idx_to_use = list(range(self.matrix.shape[0]))
+        idx_unusable = (
+            set()
+        )  # indices for which powders contain components that are not present in the target
         for component_idx in np.where(target == 0)[0]:
             present_in_powder = np.where(self.matrix[:, component_idx] > 0)[0]
             for powder_idx in present_in_powder:
-                idx_to_use.remove(powder_idx)
+                idx_unusable.add(powder_idx)
+        idx_to_use = [i for i in range(self.matrix.shape[0]) if i not in idx_unusable]
         return idx_to_use
 
     def _lookup_powder(self, s: str):
